from aiohttp import web
from multidict import istr
from typing import List
from heaserver.service import appproperty, requestproperty
from aiohttp_remotes import XForwardedRelaxed
from aiohttp_remotes.exceptions import TooManyHeaders


@web.middleware
async def new_wstl_builder(request: web.Request, handler) -> web.Response:
    wstl_builder_factory = request.app[appproperty.HEA_WSTL_BUILDER_FACTORY]
    request[requestproperty.HEA_WSTL_BUILDER] = wstl_builder_factory()
    return await handler(request)


@web.middleware
async def x_forwarded_prefix(request: web.Request, handler) -> web.Response:
    X_FORWARDED_PREFIX = istr('X-Forwarded-Prefix')
    forwarded_prefix: List[str] = request.headers.getall(X_FORWARDED_PREFIX, [])
    if forwarded_prefix:
        if len(forwarded_prefix) > 1:
            raise TooManyHeaders(X_FORWARDED_PREFIX)
        forwarded_prefix = [p.strip() for p in forwarded_prefix[0].split(',')]
        if prefix := (forwarded_prefix[0] if forwarded_prefix else None):
            request = request.clone(rel_url=_prepend_path(prefix, str(request.rel_url)))
    return await handler(request)


def _prepend_path(prefix: str, path: str) -> str:
    return prefix.rstrip('/') + '/' + path.lstrip('/')


def new_app() -> web.Application:
    """
    Creates and returns an aiohttp Application object. Installs middleware that sets the HEA_WSTL_BUILDER request
    property, assuming that the HEA_WSTL_BUILDER_FACTORY app property has already been set.

    :return: the Application property.
    """
    return web.Application(middlewares=[new_wstl_builder, x_forwarded_prefix, XForwardedRelaxed(0).middleware])




from unittest import TestCase
from heaserver.service.db.mongo import Mongo
from heaserver.service.testcase.mockmongo import MockMongo
from heaserver.service.util import public_methods


class TestMockMongo(TestCase):
    def test_has_same_public_methods(self):
        MockMongo_public_methods = public_methods(MockMongo)
        Mongo_public_methods = public_methods(Mongo)
        blacklist = [Mongo.aggregate.__name__]
        def missing_methods():
            return ', '.join(method for method in Mongo_public_methods \
                             if method not in MockMongo_public_methods and method not in blacklist)
        self.assertTrue(set(MockMongo_public_methods).issuperset(set(Mongo_public_methods) - set(blacklist)),
                        f'Missing methods: {missing_methods()}')




"""
Creates a test case class for use with the unittest library that is build into Python.
"""
import heaserver.service.db.database
import heaserver.service.testcase.mockmongo
from heaserver.service.testcase.collection import CollectionKey, query_fixture_collection
from heaserver.service.testcase.microservicetestcase import get_test_case_cls_default
from heaserver.service.testcase import expectedvalues, TEST_USER
from heaserver.service.testcase.mockmongo import MockMongoManager
from . import service
from heaobject.user import NONE_USER, ALL_USERS
from heaobject.root import Permission, HEAObjectDict
from heaobject.volume import DEFAULT_FILE_SYSTEM
from typing import Dict, List
from datetime import datetime, timezone

##
## The first set of desktop objects is for testing the following:
##
## 1) Running as the TEST_USER user, with shares that grant the test user access.
## 2) The heaserver.service.testcase.collection.CollectionKey class.
##

fixtures1: dict[str | CollectionKey, list[HEAObjectDict]] = {
    CollectionKey(name=service.MONGODB_COMPONENT_COLLECTION): [{
        'id': '666f6f2d6261722d71757578',
        'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'derived_by': None,
        'derived_from': ['foo', 'bar'],
        'description': None,
        'display_name': 'Reximus',
        'invites': [],
        'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'name': 'reximus',
        'owner': NONE_USER,
        'shares': [{
            'type': 'heaobject.root.ShareImpl',
            'invite': None,
            'user': ALL_USERS,
            'permissions': [Permission.COOWNER.name]
        }],
        'source': None,
        'source_detail': None,
        'type': 'heaobject.registry.Component',
        'base_url': 'http://localhost',
        'external_base_url': None,
        'resources': [{
            'type': 'heaobject.registry.Resource',
            'resource_type_name': 'heaobject.folder.Folder',
            'base_path': 'folders',
            'file_system_name': DEFAULT_FILE_SYSTEM,
            'file_system_type': 'heaobject.volume.MongoDBFileSystem',
            'resource_collection_type_display_name': 'heaobject.folder.Folder',
            'collection_accessor_users': [],
            'creator_users': [],
            'default_shares': []
        },
            {
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Item',
                'base_path': 'items',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Item',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            }
        ]
    },
        {
            'id': '0123456789ab0123456789ab',
            'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'derived_by': None,
            'derived_from': ['oof', 'rab'],
            'description': None,
            'display_name': 'Luximus',
            'invites': [],
            'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'name': 'luximus',
            'owner': NONE_USER,
            'shares': [{
                'type': 'heaobject.root.ShareImpl',
                'invite': None,
                'user': ALL_USERS,
                'permissions': [Permission.COOWNER.name]
            }],
            'source': None,
            'source_detail': None,
            'type': 'heaobject.registry.Component',
            'base_url': 'http://localhost/foo',
            'external_base_url': None,
            'resources': [{
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Folder',
                'base_path': 'folders',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Folder',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            },
                {
                    'type': 'heaobject.registry.Resource',
                    'resource_type_name': 'heaobject.folder.Item',
                    'base_path': 'items',
                    'file_system_name': DEFAULT_FILE_SYSTEM,
                    'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                    'resource_collection_type_display_name': 'heaobject.folder.Item',
                    'collection_accessor_users': [],
                    'creator_users': [],
                    'default_shares': []
                }]
        }
    ]}

content1 = {
    CollectionKey(name=service.MONGODB_COMPONENT_COLLECTION): {
        '666f6f2d6261722d71757578': b'The quick brown fox jumps over the lazy dog'
    }
}


def _test_case_generator(coll, fixtures, duplicate_action_name, content=None, sub=NONE_USER):
    return get_test_case_cls_default(coll=coll, fixtures=fixtures, duplicate_action_name=duplicate_action_name,
                                     db_manager_cls=heaserver.service.testcase.mockmongo.MockMongoManager,
                                     wstl_package=service.__package__, content=content, content_type='text/plain',
                                     put_content_status=204,
                                     href='http://localhost:8080/components',
                                     get_actions=[expectedvalues.Action(name='component-get-properties',
                                                                        rel=['hea-properties']),
                                                  expectedvalues.Action(name='component-get-open-choices',
                                                                        url='http://localhost:8080/components/{id}/opener',
                                                                        rel=['hea-opener-choices']),
                                                  expectedvalues.Action(name='component-duplicate',
                                                                        url='http://localhost:8080/components/{id}/duplicator',
                                                                        rel=['hea-duplicator'])],
                                     get_all_actions=[
                                         expectedvalues.Action(name='component-get-properties',
                                                               rel=['hea-properties']),
                                         expectedvalues.Action(name='component-get-open-choices',
                                                               url='http://localhost:8080/components/{id}/opener',
                                                               rel=['hea-opener-choices']),
                                         expectedvalues.Action(name='component-duplicate',
                                                               url='http://localhost:8080/components/{id}/duplicator',
                                                               rel=['hea-duplicator'])],
                                     expected_opener=expectedvalues.Link(
                                         url=f'http://localhost:8080/components/{query_fixture_collection(fixtures, key=service.MONGODB_COMPONENT_COLLECTION, default_db_manager=MockMongoManager)[0]["id"]}/content',
                                         rel=['hea-default', 'hea-opener', 'text/plain']), sub=sub)


ComponentTestCase = _test_case_generator(coll=service.MONGODB_COMPONENT_COLLECTION, fixtures=fixtures1, content=content1,
                                         duplicate_action_name='component-duplicate-form', sub=TEST_USER)


##
## The second set of desktop objects is for testing the following:
##
## 1) Running as the default (NONE_USER) user, with objects that are owned by NONE_USER.
##

fixtures2: Dict[str, List[HEAObjectDict]] = {
    service.MONGODB_COMPONENT_COLLECTION: [{
        'id': '666f6f2d6261722d71757578',
        'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'derived_by': None,
        'derived_from': ['foo', 'bar'],
        'description': None,
        'display_name': 'Reximus',
        'invites': [],
        'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'name': 'reximus',
        'owner': NONE_USER,
        'shares': [],
        'source': None,
        'source_detail': None,
        'type': 'heaobject.registry.Component',
        'base_url': 'http://localhost',
        'external_base_url': None,
        'resources': [{
            'type': 'heaobject.registry.Resource',
            'resource_type_name': 'heaobject.folder.Folder',
            'base_path': 'folders',
            'file_system_name': DEFAULT_FILE_SYSTEM,
            'file_system_type': 'heaobject.volume.MongoDBFileSystem',
            'resource_collection_type_display_name': 'heaobject.folder.Folder',
            'collection_accessor_users': [],
            'creator_users': [],
            'default_shares': []
        },
            {
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Item',
                'base_path': 'items',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Item',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            }
        ]
    },
        {
            'id': '0123456789ab0123456789ab',
            'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'derived_by': None,
            'derived_from': ['oof', 'rab'],
            'description': None,
            'display_name': 'Luximus',
            'invites': [],
            'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'name': 'luximus',
            'owner': NONE_USER,
            'shares': [],
            'source': None,
            'source_detail': None,
            'type': 'heaobject.registry.Component',
            'base_url': 'http://localhost/foo',
            'external_base_url': None,
            'resources': [{
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Folder',
                'base_path': 'folders',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Folder',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            },
                {
                    'type': 'heaobject.registry.Resource',
                    'resource_type_name': 'heaobject.folder.Item',
                    'base_path': 'items',
                    'file_system_name': DEFAULT_FILE_SYSTEM,
                    'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                    'resource_collection_type_display_name': 'heaobject.folder.Item',
                    'collection_accessor_users': [],
                    'creator_users': [],
                    'default_shares': []
                }]
        }
    ]}

content2 = {
    service.MONGODB_COMPONENT_COLLECTION: {
        '666f6f2d6261722d71757578': b'The quick brown fox jumps over the lazy dog'
    }
}

ComponentTestCase2 = _test_case_generator(coll=service.MONGODB_COMPONENT_COLLECTION, fixtures=fixtures2,
                                          duplicate_action_name='component-duplicate-form', content=content2)

##
## The third set of desktop objects is for testing the following:
##
## 1) Proper handling of an omitted list property (shares, in this case).
##

fixtures3: Dict[str, List[HEAObjectDict]] = {
    service.MONGODB_COMPONENT_COLLECTION: [{
        'id': '666f6f2d6261722d71757578',
        'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'derived_by': None,
        'derived_from': ['foo', 'bar'],
        'description': None,
        'display_name': 'Reximus',
        'invites': [],
        'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
        'name': 'reximus',
        'owner': NONE_USER,
        'source': None,
        'source_detail': None,
        'shares': [],
        'type': 'heaobject.registry.Component',
        'base_url': 'http://localhost',
        'external_base_url': None,
        'resources': [{
            'type': 'heaobject.registry.Resource',
            'resource_type_name': 'heaobject.folder.Folder',
            'base_path': 'folders',
            'file_system_name': DEFAULT_FILE_SYSTEM,
            'file_system_type': 'heaobject.volume.MongoDBFileSystem',
            'resource_collection_type_display_name': 'heaobject.folder.Folder',
            'collection_accessor_users': [],
            'creator_users': [],
            'default_shares': []
        },
            {
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Item',
                'base_path': 'items',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Item',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            }
        ]
    },
        {
            'id': '0123456789ab0123456789ab',
            'created': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'derived_by': None,
            'derived_from': ['oof', 'rab'],
            'description': None,
            'display_name': 'Luximus',
            'invites': [],
            'modified': datetime(2021, 12, 2, 17, 31, 15, 630000, tzinfo=timezone.utc),
            'name': 'luximus',
            'owner': NONE_USER,
            'shares': [],
            'source': None,
            'source_detail': None,
            'type': 'heaobject.registry.Component',
            'base_url': 'http://localhost/foo',
            'external_base_url': None,
            'resources': [{
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Folder',
                'base_path': 'folders',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Folder',
                'collection_accessor_users': [],
                'creator_users': [],
                'default_shares': []
            },
                {
                    'type': 'heaobject.registry.Resource',
                    'resource_type_name': 'heaobject.folder.Item',
                    'base_path': 'items',
                    'file_system_name': DEFAULT_FILE_SYSTEM,
                    'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                    'resource_collection_type_display_name': 'heaobject.folder.Item',
                    'collection_accessor_users': [],
                    'creator_users': [],
                    'default_shares': []
                }]
        }
    ]}

content3 = {
    service.MONGODB_COMPONENT_COLLECTION: {
        '666f6f2d6261722d71757578': b'The quick brown fox jumps over the lazy dog'
    }
}

ComponentTestCase3 = _test_case_generator(coll=service.MONGODB_COMPONENT_COLLECTION, fixtures=fixtures3,
                                          duplicate_action_name='component-duplicate-form', content=content3)



from unittest import IsolatedAsyncioTestCase
from .awss3foldertestcase import AWSS3FolderTestCase
from heaserver.service.db import awsservicelib
from heaserver.service.db.aws import S3
from heaobject.awss3key import encode_key
import boto3
from botocore.exceptions import ClientError
from aiohttp import hdrs
from freezegun.api import FakeDatetime
from dateutil.tz import tzutc
from heaobject.user import NONE_USER
from moto import mock_sts


class TestS3GetAccount(IsolatedAsyncioTestCase):

    @mock_sts()
    def run(self, result):
        super().run(result)

    async def test_get_account(self):
        sts = boto3.client('sts')
        actual = await S3._get_basic_account_info(sts)
        self.assertEquals('123456789012', actual.id)


class TestAWSProperties(AWSS3FolderTestCase):
    async def test_get_property(self):
        async with self.client.request('GET', '/properties/CLOUD_AWS_CRED_URL') as resp:
            self.assertEquals(200, resp.status)

    async def test_get_property_not_found(self):
        async with self.client.request('GET', '/properties/TEST') as resp:
            self.assertEquals(404, resp.status)


class TestMockAWSAccount(AWSS3FolderTestCase):
    @mock_sts()
    def run(self, result):
        super().run(result)

    async def test_get_volume_id_for_account_id(self):
        async with self.client.request('GET', '/accounts/123456789012') as resp:
            self.assertEquals(200, resp.status)


class TestAWSServiceLib(AWSS3FolderTestCase):

    def setUp(self):
        super().setUp()
        self.s3 = boto3.client('s3')

    async def test_list_bucket_not_found(self):
        with self.assertRaises(ClientError):
            l = [o async for o in awsservicelib.list_objects(self.s3, 'blah')]

    async def test_list_bucket(self):
        expected = [{'Key': 'TestFolder/', 'LastModified': FakeDatetime(2022, 5, 17, 0, 0, tzinfo=tzutc()),
                     'ETag': '"d41d8cd98f00b204e9800998ecf8427e"', 'Size': 0, 'StorageClass': 'STANDARD'},
                    {'Key': 'TestFolder2/', 'LastModified': FakeDatetime(2022, 5, 17, 0, 0, tzinfo=tzutc()),
                     'ETag': '"d41d8cd98f00b204e9800998ecf8427e"', 'Size': 0, 'StorageClass': 'STANDARD'}]
        actual = [o async for o in awsservicelib.list_objects(self.s3, 'arp-scale-2-cloud-bucket-with-tags11')]
        self.assertEquals(expected, actual)

    async def test_list_empty_bucket_with_filter(self):
        expected = [{'Key': 'TestFolder/', 'LastModified': FakeDatetime(2022, 5, 17, 0, 0, tzinfo=tzutc()),
                     'ETag': '"d41d8cd98f00b204e9800998ecf8427e"', 'Size': 0, 'StorageClass': 'STANDARD'},
                    {'Key': 'TestFolder2/', 'LastModified': FakeDatetime(2022, 5, 17, 0, 0, tzinfo=tzutc()),
                     'ETag': '"d41d8cd98f00b204e9800998ecf8427e"', 'Size': 0, 'StorageClass': 'STANDARD'}]
        actual = [o async for o in
                  awsservicelib.list_objects(self.s3, 'arp-scale-2-cloud-bucket-with-tags11', prefix='TestFolder')]
        self.assertEquals(expected, actual)

    async def test_list_empty_bucket_with_filter_one(self):
        expected = [{'Key': 'TestFolder/', 'LastModified': FakeDatetime(2022, 5, 17, 0, 0, tzinfo=tzutc()),
                     'ETag': '"d41d8cd98f00b204e9800998ecf8427e"', 'Size': 0, 'StorageClass': 'STANDARD'}]
        actual = [o async for o in
                  awsservicelib.list_objects(self.s3, 'arp-scale-2-cloud-bucket-with-tags11', prefix='TestFolder/')]
        self.assertEquals(expected, actual)

    async def test_delete_object_status(self):
        actual = await awsservicelib._delete_object(self.s3, bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                    key='TestFolder2/', recursive=True)
        self.assertEquals(204, actual.status)

    async def test_delete_object(self):
        await awsservicelib._delete_object(self.s3, bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                           key='TestFolder2/', recursive=True)
        self.assertFalse(await awsservicelib._object_exists(self.s3,
                                                            'arp-scale-2-cloud-bucket-with-tags11',
                                                            'TestFolder2/'))

    async def test_delete_object_not_found_status(self):
        actual = await awsservicelib._delete_object(self.s3, bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                    key='foobar/', recursive=True)
        self.assertEquals(404, actual.status)

    async def test_object_not_found_status(self):
        self.assertFalse(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags11', 'foobar/'))

    async def test_delete_object_bucket_not_found(self):
        actual = await awsservicelib._delete_object(self.s3, bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                    key='TestFolder2/', recursive=True)
        self.assertEquals(404, actual.status)

    async def test_copy_object_status(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder2/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='TestFolder/')
        self.assertEquals(201, actual.status, actual.text)

    async def test_copy_object(self):
        await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         source_key='TestFolder2/',
                                         target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         target_key='TestFolder/')
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags11',
                                                           'TestFolder/TestFolder2/'))

    async def test_copy_folder_into_itself(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='TestFolder/')
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_object_recursive(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder2/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='TestFolder/')
        if actual.status != 201:
            self.fail(f'1: {actual.text}')
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='TestFolder2/')
        if actual.status != 201:
            self.fail(f'2: {actual.text}')
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags11',
                                                           'TestFolder2/TestFolder/TestFolder2/'))

    async def test_copy_object_to_different_bucket_status(self):
        resp = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                source_key='TestFolder2/',
                                                target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                target_key='')
        self.assertEquals(201, resp.status, resp.text)

    async def test_copy_object_to_different_bucket_at_root(self):
        await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         source_key='TestFolder2/',
                                         target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                         target_key='')
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags1',
                                                           'TestFolder2/'))

    async def test_copy_object_to_different_bucket_at_root2(self):
        await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         source_key='TestFolder2/',
                                         target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                         target_key=None)
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags1',
                                                           'TestFolder2/'))

    async def test_copy_whole_bucket(self):
        await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         source_key='',
                                         target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                         target_key='')
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags1',
                                                           'TestFolder2/'))

    async def test_copy_whole_bucket_none(self):
        await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                         source_key=None,
                                         target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                         target_key=None)
        self.assertTrue(await awsservicelib._object_exists(self.s3, 'arp-scale-2-cloud-bucket-with-tags1',
                                                           'TestFolder2/'))

    async def test_copy_whole_bucket_same_bucket(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key=None,
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key=None)
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_whole_bucket_empty_same_bucket(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                  source_key=None,
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                  target_key=None)
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_object_not_found_source_bucket(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                  source_key='TestFolder2/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='foobar/TestFolder3/')
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_object_not_found_destination_bucket(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder2/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags1',
                                                  target_key='foobar/TestFolder3/')
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_object_not_found_object(self):
        actual = await awsservicelib._copy_object(self.s3, source_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  source_key='TestFolder22/',
                                                  target_bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                  target_key='foobar/TestFolder3/')
        self.assertEquals(400, actual.status, actual.text)

    async def test_copy_rest_status(self):
        href = f'/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder/")}/duplicator'
        body = {'template':
            {'data': [
                {'name': 'target',
                 'value': f'http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder2/")}'}]}}
        async with self.client.post(href, json=body) as resp:
            self.assertEquals(201, resp.status, await resp.text())

    async def test_copy_rest_header(self):
        href = f'/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder/")}/duplicator'
        body = {'template':
            {'data': [
                {'name': 'target',
                 'value': f'http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder2/")}'}]}}
        async with self.client.post(href, json=body) as resp:
            self.assertEquals(
                f'http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/{encode_key("TestFolder2/")}',
                resp.headers.get(hdrs.LOCATION), await resp.text())

    async def test_create_folder_status(self):
        actual = await awsservicelib._create_object(self.s3,
                                                    bucket_name='arp-scale-2-cloud-bucket-with-tags11',
                                                    key='TestFolder2/TestFolder/')
        self.assertEqual(201, actual.status, actual.text)

    async def test_create_folder_bad_bucket_name(self):
        actual = await awsservicelib._create_object(self.s3,
                                                    bucket_name='arp-scale-2-cloud-bucket-with-tags11-bad',
                                                    key='TestFolder2/TestFolder/')
        self.assertEqual(404, actual.status, actual.text)

    async def test_create_folder_no_bucket(self):
        with self.assertRaises(ValueError):
            await awsservicelib._create_object(self.s3,
                                               bucket_name=None,
                                               key='TestFolder2/TestFolder/')

    async def test_create_folder_bucket_empty_string(self):
        actual = await awsservicelib._create_object(self.s3,
                                                    bucket_name='',
                                                    key='TestFolder2/TestFolder/')
        self.assertEqual(400, actual.status, actual.text)

    async def test_create_folder_no_key(self):
        with self.assertRaises(ValueError):
            await awsservicelib._create_object(self.s3,
                                               bucket_name='arp-scale-2-cloud-bucket-with-tags11-bad',
                                               key=None)

    async def test_create_folder_key_empty_string(self):
        actual = await awsservicelib._create_object(self.s3,
                                                    bucket_name='arp-scale-2-cloud-bucket-with-tags11-bad',
                                                    key='')
        self.assertEqual(400, actual.status, actual.text)

    async def test_create_folder_rest_status(self):
        href = f'/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/'
        body = {'bucket_id': 'arp-scale-2-cloud-bucket-with-tags11', 'created': '2022-05-17T00:00:00+00:00',
                'derived_by': None, 'derived_from': [], 'description': None, 'display_name': 'TestFolder',
                'id': 'VGVzdEZvbGRlcjIvVGVzdEZvbGRlci8=', 'invites': [], 'is_folder': True,
                'key': 'TestFolder2/TestFolder/', 'mime_type': 'application/x.folder',
                'modified': '2022-05-17T00:00:00+00:00', 'name': 'VGVzdEZvbGRlcjIvVGVzdEZvbGRlci8=',
                'owner': 'system|none',
                'path': '/arp-scale-2-cloud-bucket-with-tags11/TestFolder2/TestFolder/', 'presigned_url': None,
                's3_uri': 's3://arp-scale-2-cloud-bucket-with-tags11/TestFolder2/TestFolder/', 'shares': [],
                'source': 'AWS S3', 'storage_class': 'STANDARD',
                'type': 'heaobject.folder.AWSS3Folder'}
        async with self.client.post(href, json=body) as resp:
            self.assertEqual(201, resp.status, await resp.text())

    async def test_create_folder_rest_header(self):
        href = f'/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/'
        body = {'bucket_id': 'arp-scale-2-cloud-bucket-with-tags11', 'created': '2022-05-17T00:00:00+00:00',
                'derived_by': None, 'derived_from': [], 'description': None, 'display_name': 'TestFolder',
                'id': 'VGVzdEZvbGRlcjIvVGVzdEZvbGRlci8=', 'invites': [], 'is_folder': True,
                'key': 'TestFolder2/TestFolder/', 'mime_type': 'application/x.folder',
                'modified': '2022-05-17T00:00:00+00:00', 'name': 'VGVzdEZvbGRlcjIvVGVzdEZvbGRlci8=',
                'owner': 'system|none',
                'path': '/arp-scale-2-cloud-bucket-with-tags11/TestFolder2/TestFolder/', 'presigned_url': None,
                's3_uri': 's3://arp-scale-2-cloud-bucket-with-tags11/TestFolder2/TestFolder/', 'shares': [],
                'source': 'AWS S3', 'storage_class': 'STANDARD',
                'type': 'heaobject.folder.AWSS3Folder'}
        async with self.client.post(href, json=body) as resp:
            if resp.status != 201:
                self.fail(await resp.text())
            else:
                self.assertEqual(
                    'http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/arp-scale-2-cloud-bucket-with-tags11/awss3folders/VGVzdEZvbGRlcjIvVGVzdEZvbGRlci8=',
                    resp.headers[hdrs.LOCATION])



from heaserver.service.testcase.simpleaiohttptestcase import SimpleAioHTTPTestCase
from heaserver.service import wstl
from heaserver.service.representor import cj, error
from heaobject import user
from datetime import datetime
from unittest.mock import AsyncMock
from aiohttp.web import Request
from aiohttp import hdrs
import json


class TestCj(SimpleAioHTTPTestCase):

    async def setUpAsync(self):
        await super().setUpAsync()
        self.__body = {
            'items': [],
            'created': datetime(2021, 12, 2, 17, 31, 15, 630000),
            'derived_by_uri': None,
            'derived_from_uris': [],
            'description': None,
            'display_name': 'Reximus',
            'id': '1',
            'invites': [],
            'modified': None,
            'name': 'reximus',
            'owner': user.NONE_USER,
            'shares': [],
            'source_uri': None,
            'type': 'heaobject.folder.Folder',
            'version': None,
            'nested_dict': {'foo': 'oof', 'baz': 'zab', 'type': 'heaobject.registry.Component'},
            'nested_list': [{'foo': 'oof', 'baz': 'zab', 'type': 'heaobject.registry.Component'},
                            {'foo': 'oof', 'baz': 'zab', 'type': 'heaobject.registry.Component'}],
            'empty_nested_list': [],
            'plain_list': [1, 2, 3, 4]
        }
        self.maxDiff = None

    async def test_formats_default(self):
        b = wstl.builder('servicetest', resource='representor/all.json')
        actual, _ = await b.represent_from_request(self.__request(), None)
        expected = '[{"collection": {"href": "#", "permissions": [], "version": "1.0"}}]'
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection":
                          {"version": "1.0",
                           "href": "http://localhost/test",
                           "permissions": [[]],
                           "items": [
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                   ],
                                   "links": []
                               }
                           ]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_two_items_empty_permissions(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), [self.__body, self.__body])
        expected = '''[{"collection":
                          {"version": "1.0",
                           "href": "http://localhost/test",
                           "permissions": [[], []],
                           "items": [
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                   ],
                                   "links": []
                               },
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                   ],
                                   "links": []
                               }
                           ]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_item_href(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_item_href.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-open', path='foo/bar/{id}',
                                                 root='http://localhost:8080')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection":
                         {"version": "1.0",
                          "href": "http://localhost/test",
                          "permissions": [[]],
                          "items": [
                              {"data": [
                                  {"name": "items", "value": [], "prompt": "items", "display": true},
                                  {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                  {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                  {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                  {"name": "description", "value": null, "prompt": "description", "display": true},
                                  {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                  {"name": "id", "value": "1", "prompt": "id", "display": false},
                                  {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                  {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                  {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                  {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                  {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                  {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                  {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                  {"name": "version", "value": null, "prompt": "version", "display": true},
                                  {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                  {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                  ],
                                  "rel": "",
                                  "href": "http://localhost:8080/foo/bar/1",
                                  "links": []
                              }
                          ]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_item_link(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_item_link.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-open', path='foo/bar/{id}',
                                                 root='http://localhost:8080')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection":
                          {"version": "1.0",
                           "href": "http://localhost/test",
                           "permissions": [[]],
                           "items": [
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                ],
                                "links": [
                                    {"prompt": "Open", "rel": "", "href": "http://localhost:8080/foo/bar/1"}
                                ]
                           }]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_item_link_item_if_True(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_item_link.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-open', path='foo/bar/{id}',
                                                 root='http://localhost:8080', itemif='True')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection":
                          {"version": "1.0",
                           "href": "http://localhost/test",
                           "permissions": [[]],
                           "items": [
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                ],
                                "links": [
                                    {"prompt": "Open", "rel": "", "href": "http://localhost:8080/foo/bar/1"}
                                ]
                           }]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_item_link_item_if_False(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_item_link.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-open', path='/foo/bar/{id}',
                                                 root='http://localhost:8080', itemif='False')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection":
                          {"version": "1.0",
                           "href": "http://localhost/test",
                           "permissions": [[]],
                           "items": [
                               {"data": [
                                   {"name": "items", "value": [], "prompt": "items", "display": true},
                                   {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                   {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                   {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                   {"name": "description", "value": null, "prompt": "description", "display": true},
                                   {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                   {"name": "id", "value": "1", "prompt": "id", "display": false},
                                   {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                   {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                   {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                   {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                   {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                   {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                   {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                   {"name": "version", "value": null, "prompt": "version", "display": true},
                                   {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                   {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                   {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                   {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                ],
                                "links": []
                           }]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_toplevel_link_for_item(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_toplevel_link_for_item.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-open', path='foo/bar', root='http://localhost:8080')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection": {
                          "version": "1.0",
                          "href": "http://localhost/test",
                          "permissions": [[]],
                          "items": [
                              {"data": [
                                  {"name": "items", "value": [], "prompt": "items", "display": true},
                                  {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                  {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                  {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                  {"name": "description", "value": null, "prompt": "description", "display": true},
                                  {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                  {"name": "id", "value": "1", "prompt": "id", "display": false},
                                  {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                  {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                  {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                  {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                  {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                  {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                  {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                  {"name": "version", "value": null, "prompt": "version", "display": true},
                                  {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                  {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                  {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                  {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                               ],
                               "links": []
                               }
                          ],
                          "links": [
                              {"prompt": "Open", "rel": "", "href": "http://localhost:8080/foo/bar"}
                          ]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_toplevel_link_for_list(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_toplevel_link_for_list.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-list', path='foo/bar', root='http://localhost/test')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection": {
                                  "version": "1.0",
                                  "href": "http://localhost/test",
                                  "permissions": [[]],
                                  "items": [
                                      {"data": [
                                          {"name": "items", "value": [], "prompt": "items", "display": true},
                                          {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                          {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                          {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                          {"name": "description", "value": null, "prompt": "description", "display": true},
                                          {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                          {"name": "id", "value": "1", "prompt": "id", "display": false},
                                          {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                          {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                          {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                          {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                          {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                          {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                          {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                          {"name": "version", "value": null, "prompt": "version", "display": true},
                                          {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                          {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                       ],
                                       "links": []
                                       }
                                  ],
                                  "links": [
                                      {"prompt": "List", "rel": "", "href": "http://localhost/test/foo/bar"}
                                  ]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_queries(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_queries.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-search', path='foo/bar', root='http://localhost/test')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection": {
                              "version": "1.0",
                              "href": "http://localhost/test",
                              "permissions": [[]],
                              "items": [
                                  {"data": [
                                      {"name": "items", "value": [], "prompt": "items", "display": true},
                                      {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                      {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                      {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                      {"name": "description", "value": null, "prompt": "description", "display": true},
                                      {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                      {"name": "id", "value": "1", "prompt": "id", "display": false},
                                      {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                      {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                      {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                      {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                      {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                      {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                      {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                      {"name": "version", "value": null, "prompt": "version", "display": true},
                                      {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                      {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                      {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                      {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                      {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                      {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                      {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                      {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                      {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                      {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                      {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                   ],
                                   "links": []
                                   }
                              ],
                              "queries": [{
                                  "data": [{
                                      "name": "keywords",
                                      "pattern": null,
                                      "prompt": "Keywords",
                                      "readOnly": false,
                                      "required": true,
                                      "value": ""}, {
                                      "name": "match-case",
                                      "pattern": null,
                                      "prompt": "Match case?",
                                      "readOnly": false,
                                      "required": true,
                                      "value": ""}],
                                  "href": "http://localhost/test/foo/bar",
                                  "prompt": "Search",
                                  "rel": ""}]}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_formats_data_with_template(self):
        runtime_wstl_builder = wstl.builder('servicetest.representor', resource='all_cj_template.json')
        runtime_wstl_builder.href = 'http://localhost/test'
        runtime_wstl_builder.add_run_time_action('data-adapter-search', path='/foo/bar', root='http://localhost/test')
        actual, _ = await runtime_wstl_builder.represent_from_request(self.__request(), self.__body)
        expected = '''[{"collection": {
                                  "version": "1.0",
                                  "href": "http://localhost/test",
                                  "permissions": [[]],
                                  "items": [
                                      {"data": [
                                          {"name": "items", "value": [], "prompt": "items", "display": true},
                                          {"name": "created", "value": "2021-12-02T17:31:15.630000", "prompt": "created", "display": true},
                                          {"name": "derived_by_uri", "value": null, "prompt": "derived_by_uri", "display": true},
                                          {"name": "derived_from_uris", "value": [], "prompt": "derived_from_uris", "display": true},
                                          {"name": "description", "value": null, "prompt": "description", "display": true},
                                          {"name": "display_name", "value": "Reximus", "prompt": "display_name", "display": true},
                                          {"name": "id", "value": "1", "prompt": "id", "display": false},
                                          {"name": "invites", "value": [], "prompt": "invites", "display": true},
                                          {"name": "modified", "value": null, "prompt": "modified", "display": true},
                                          {"name": "name", "value": "reximus", "prompt": "name", "display": true},
                                          {"name": "owner", "value": "system|none", "prompt": "owner", "display": true},
                                          {"name": "shares", "value": [], "prompt": "shares", "display": true},
                                          {"name": "source_uri", "value": null, "prompt": "source_uri", "display": true},
                                          {"name": "type", "value": "heaobject.folder.Folder", "prompt": "type", "display": true},
                                          {"name": "version", "value": null, "prompt": "version", "display": true},
                                          {"section": "nested_dict", "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_dict", "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_dict", "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_list", "index": 0, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "foo", "value": "oof", "prompt": "foo", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "baz", "value": "zab", "prompt": "baz", "display": true},
                                          {"section": "nested_list", "index": 1, "name": "type", "value": "heaobject.registry.Component", "prompt": "type", "display": true},
                                          {"name": "empty_nested_list", "prompt": "empty_nested_list", "value": [], "display": true},
                                          {"name": "plain_list", "prompt": "plain_list", "value": [1, 2, 3, 4], "display": true}
                                       ],
                                       "links": []
                                       }
                                  ],
                                  "template": {
                                      "data": [
                                          {"name": "keywords",
                                           "pattern": null,
                                           "prompt": "Keywords",
                                           "readOnly": false,
                                           "required": true,
                                           "value": ""},
                                          {"name": "match-case",
                                           "pattern": null,
                                           "prompt": "Match case?",
                                           "readOnly": false,
                                           "required": true,
                                           "value": "",
                                           "type": "select",
                                           "options": [{"value": "true"}, {"value": "false"}]},
                                           {"section": "nested_dict",
                                           "name": "foo",
                                           "pattern": null,
                                           "prompt": "foo",
                                           "required": true,
                                           "value": "oof",
                                           "readOnly": false},
                                           {"section": "nested_dict",
                                           "name": "baz",
                                           "pattern": null,
                                           "prompt": "baz",
                                           "required": true,
                                           "value": "zab",
                                           "readOnly": false},
                                           {"name": "empty_nested_list",
                                           "pattern": null,
                                           "prompt": "Empty nested list",
                                           "readOnly": false,
                                           "required": true,
                                           "value": []},
                                           {"section": "nested_list",
                                           "index": -1,
                                           "name": "foo",
                                           "pattern": null,
                                           "prompt": "foo",
                                           "required": true,
                                           "value": "",
                                           "readOnly": false},
                                           {"section": "nested_list",
                                           "index": -1,
                                           "name": "baz",
                                           "pattern": null,
                                           "prompt": "baz",
                                           "required": true,
                                           "value": "",
                                           "readOnly": false},
                                           {"section": "nested_list",
                                           "index": 0,
                                           "name": "foo",
                                           "pattern": null,
                                           "prompt": "foo",
                                           "required": true,
                                           "value": "oof",
                                           "readOnly": false},
                                           {"section": "nested_list",
                                           "index": 0,
                                           "name": "baz",
                                           "pattern": null,
                                           "prompt": "baz",
                                           "required": true,
                                           "value": "zab",
                                           "readOnly": false},
                                           {"section": "nested_list",
                                           "index": 1,
                                           "name": "foo",
                                           "pattern": null,
                                           "prompt": "foo",
                                           "required": true,
                                           "value": "oof",
                                           "readOnly": false},
                                           {"section": "nested_list",
                                           "index": 1,
                                           "name": "baz",
                                           "pattern": null,
                                           "prompt": "baz",
                                           "required": true,
                                           "value": "zab",
                                           "readOnly": false},
                                           {"name": "plain_list",
                                           "pattern": null,
                                           "prompt": "Plain list",
                                           "required": true,
                                           "value": [1, 2, 3, 4],
                                           "readOnly": false}],
                                      "prompt": "Search",
                                      "rel": ""}}}]'''
        self._assert_json_string_equals(expected, actual)

    async def test_parses_template(self):
        cj_str = '{"template": {"data": [{"name": "foo", "value": "bar"}, {"name": "baz", "value": "oof"}]}}'
        actual = await cj.CJ().parse(self.__request(cj_str))
        expected = {"foo": "bar", "baz": "oof"}
        self.assertEqual(expected, actual)

    async def test_parses_missing_data_object(self):
        cj_str = '{"template": {}}'
        actual = await cj.CJ().parse(self.__request(cj_str))
        expected = {}
        self.assertEqual(expected, actual)

    async def test_parses_empty_data_array(self):
        cj_str = '{"template": {"data": []}}'
        actual = await cj.CJ().parse(self.__request(cj_str))
        expected = {}
        self.assertEqual(expected, actual)

    async def test_parses_missing_template_property(self):
        with self.assertRaises(error.ParseException):
            await cj.CJ().parse(self.__request('{"tmplte": {"data": []}}'))

    async def test_parses_empty(self):
        with self.assertRaises(error.ParseException):
            await cj.CJ().parse(self.__request('{}'))

    async def test_parses_no_name(self):
        with self.assertRaises(error.ParseException):
            await cj.CJ().parse(self.__request('{"template": {"data": [{"nm": "foo"}]}}'))

    async def test_parses_no_list(self):
        with self.assertRaises(error.ParseException):
            await cj.CJ().parse(self.__request('{"template": {"data": {"name": "foo"}}}'))

    def __request(self, jsons: str = None) -> AsyncMock:
        request = AsyncMock(spec=Request)
        request.text.return_value = jsons
        request.json.side_effect = lambda loads=json.loads: loads(jsons)
        request.match_info = {}
        request.headers = {hdrs.ACCEPT: cj.MIME_TYPE}
        return request

    def test_to_nvpjson(self):
        input = {'template': {'data': [
            {"name": "id",
             "value": "1"},
            {"name": "derived_by_uri",
             "value": None},
            {"name": "derived_from_uris",
             "value": []},
            {"name": "invites",
             "value": []},
            {"name": "description",
             "value": None},
            {"name": "modified",
             "value": None},
            {"name": "source_uri",
             "value": None},
            {"name": "version",
             "value": None},
            {"name": "display_name",
             "value": "Reximus"},
            {"name": "name",
             "value": "reximus"},
            {"name": "owner",
             "value": user.NONE_USER},
            {"section": "nested_dict",
             "name": "foo",
             "value": "oof"},
            {"section": "nested_dict",
             "name": "baz",
             "value": "zab"},
            {"name": "empty_nested_list",
             "value": []},
            {"name": "shares",
             "value": []},
            {"section": "nested_list",
             "index": 0,
             "name": "foo",
             "value": "oof"},
            {"section": "nested_list",
             "index": 1,
             "name": "foo",
             "value": "oof"},
            {"section": "nested_list",
             "index": 0,
             "name": "baz",
             "value": "zab"},
            {"section": "nested_list",
             "index": 1,
             "name": "baz",
             "value": "zab"},
            {"name": "plain_list",
             "value": [1, 2, 3, 4]},
            {"name": "created",
             "value": datetime(2021, 12, 2, 17, 31, 15, 630000)}
        ]}}
        expected = {
            'created': datetime(2021, 12, 2, 17, 31, 15, 630000),
            'derived_by_uri': None,
            'derived_from_uris': [],
            'description': None,
            'display_name': 'Reximus',
            'id': '1',
            'invites': [],
            'modified': None,
            'name': 'reximus',
            'owner': user.NONE_USER,
            'shares': [],
            'source_uri': None,
            'version': None,
            'nested_dict': {'foo': 'oof', 'baz': 'zab'},
            'nested_list': [{'foo': 'oof', 'baz': 'zab'},
                            {'foo': 'oof', 'baz': 'zab'}],
            'empty_nested_list': [],
            'plain_list': [1, 2, 3, 4]
        }
        actual = cj.to_nvpjson(input)
        return self.assertEqual(expected, actual)

    def test_add_extended_property_values_with_optionsFromUrl(self):
        input = {
            "hea": {
                "optionsFromUrl": {
                    'href': 'http://localhost:8080/people/',
                    "path": "/people/",
                    "value": "name",
                    "text": "display_name"
                }
            },
            "type": "select",
            "name": "user",
            "prompt": "User"
        }
        actual = {}
        expected = {
            'options': {
                'href': 'http://localhost:8080/people/',
                'text': 'display_name',
                'value': 'name'
            },
            "type": "select"
        }
        cj.add_extended_property_values(input, actual)
        self.assertEqual(expected, actual)

    def test_add_extended_property_values_with_suggest(self):
        input = {
            "suggest": [
                {"value": "COOWNER", "text": "Co-owner"},
                {"value": "CREATOR", "text": "Creator"},
                {"value": "DELETER", "text": "Deleter"},
                {"value": "EDITOR", "text": "Editor"},
                {"value": "SHARER", "text": "Sharer"},
                {"value": "VIEWER", "text": "Viewer"}
            ],
            "type": "select",
            "name": "user",
            "prompt": "User"
        }
        actual = {}
        expected = {
            'options': [
                {"value": "COOWNER", "text": "Co-owner"},
                {"value": "CREATOR", "text": "Creator"},
                {"value": "DELETER", "text": "Deleter"},
                {"value": "EDITOR", "text": "Editor"},
                {"value": "SHARER", "text": "Sharer"},
                {"value": "VIEWER", "text": "Viewer"}
            ],
            "type": "select"
        }
        cj.add_extended_property_values(input, actual)
        self.assertEqual(expected, actual)

    def test_add_extended_property_values_with_related(self):
        input = {
            "suggest": {
                "related": "testList",
                "value": "id",
                "text": "userName"
            },
            "type": "select",
            "name": "user",
            "prompt": "User",
            "related": {
                "testList": [
                    {
                        "id": 1,
                        "userName": 'joe'
                    }
                ]
            }
        }
        actual = {}
        expected = {
            'options': [{'text': 'joe', 'value': 1}], 'type': 'select'
        }
        cj.add_extended_property_values(input, actual)
        self.assertEqual(expected, actual)



from .aiohttptestcase import HEAAioHTTPTestCase
from .collection import CollectionKey, validate_collection_keys, convert_to_collection_keys
from .expectedvalues import Action, Link, expected_values
from .. import wstl
from aiohttp.web import Application
from ..oidcclaimhdrs import SUB
from .testenv import app_context, RegistryContainerConfig, DockerContainerConfig
from ..db.database import MicroserviceDatabaseManager
from heaserver.service.testcase.mockmongo import MockMongoManager
from aiohttp import hdrs
from heaobject.user import NONE_USER
from heaobject.root import DesktopObjectDict
from typing import Optional, Any, Iterable, Type, Mapping
from .util import maybe_docker_is_not_running
from copy import deepcopy
from yarl import URL
import logging


class MicroserviceTestCase(HEAAioHTTPTestCase):
    """
    Abstract base class for testing HEA microservices. It provides many expected values that should be returned from
    requests.
    """

    def __init__(self,
                 coll: str,
                 desktop_objects: Mapping[str, list[DesktopObjectDict]] | Mapping[
                     CollectionKey, list[DesktopObjectDict]],
                 db_manager_cls: Optional[Type[MicroserviceDatabaseManager]] = MockMongoManager,
                 wstl_package: str | None = None,
                 href: Optional[str | URL] = None,
                 body_post: Optional[Mapping[str, Mapping[str, Any]]] = None,
                 body_put: Optional[Mapping[str, Mapping[str, Any]]] = None,
                 content_id: Optional[str] = None,
                 expected_all: Optional[list[Mapping[str, Any]]] = None,
                 expected_one: Optional[list[Mapping[str, Any]]] = None,
                 expected_one_wstl: Optional[Mapping[str, Any]] = None,
                 expected_all_wstl: Optional[Mapping[str, Any]] = None,
                 expected_one_duplicate_form: Optional[Mapping[str, Any]] = None,
                 expected_opener: Optional[str | URL] = None,
                 expected_opener_body: Optional[Mapping[str, Any]] = None,
                 content: Mapping[str, Mapping[str, bytes]] | Mapping[CollectionKey, Mapping[str, bytes]] | None = None,
                 content_type: Optional[str] = None,
                 put_content_status: Optional[int] = None,
                 methodName='runTest',
                 port: Optional[int] = None,
                 sub: Optional[str] = NONE_USER,
                 registry_docker_image: Optional[RegistryContainerConfig] = None,
                 other_docker_images: Iterable[DockerContainerConfig] | None = None,
                 package_name: str | None = None):
        """
        Initializes a test case.

        :param coll: collection name (required).
        :param desktop_objects: HEA desktop objects to load into the database (required), as a map of collection ->
        list of desktop object dicts.
        :param db_manager_cls: the database factory class. Defaults to TestMockMongoManager.
        :param wstl_package: the name of the package containing the WeSTL data package.
        :param href: the resource being tested. If None, uses http://localhost:{port}/{coll}/, where {coll} is the
        collection name, and {port} is the port ultimately used by this test case.
        :param body_post: a Collection+JSON template as a dictionary, to be used for submitting POST requests.
        Populates the _body_post protected attribute. If None, _body_post will be None.
        :param body_put: a Collection+JSON template as a dictionary, to be used for submitting PUT requests. Populates
        the _body_put protected attribute. If None, _body_put will be None. Populates the _body_put protected
        attribute. If None, _body_put will be None.
        :param content_id: the ID of an object that has content, to be used for getting and updating content.
        :param expected_all: the expected JSON dictionary list from a GET all request for a Collection+JSON document.
        Populates the _expected_all protected attribute. If None, _expected_all will be None.
        :param expected_one: the expected JSON dictionary list from a GET one request for a Collection+JSON document.
        If None, expected_all will be used. Populates the _expected_one attribute. If None and expected_all is None,
        _expected_one will be None.
        :param expected_one_wstl: the expected JSON dictionary list from a GET one request for a run-time WeSTL
        document. Populates the _expected_one_wstl protected attribute. If None, the value of expected_all_wstl will be
        used. If None and expected_all_wstl is None, _expected_one_wstl will be None.
        :param expected_all_wstl: the expected JSON dictionary list from a GET all request for a run-time WeSTL
        document. Populates the _expected_all_wstl protected attribute. If None, _expected_all_wstl will be None.
        :param expected_one_duplicate_form: expected JSON from a GET request for a form template to duplicate an
        object. Populates the _expected_one_duplicate_form protected attribute. If None, _expected_one_duplicate_form
        will be None.
        :param expected_opener: the expected URL of the resource that does the opening.
        :param expected_opener_body: the expected JSON dictionary for GET calls for an HEA desktop object's opener
        choices.
        :param content: content to send via a PUT content request. It should not be the same as the content already
        associated with the object whose ID is content_id.
        :param content_type: the MIME type of the content included with the objects (e.g., text/plain; charset=utf-8).
        Must be specified if content_type is specified.
        :param put_content_status: the status code expected from a response to a PUT content request.
        :param methodName: the name of the method to test.
        :param port: the port number to run aiohttp. If None, a random available port will be chosen.
        :param sub: the user from which the requests will be sent. Defaults to heaobject.user.NONE_USER.
        :param registry_docker_image: a RegistryContainerConfig for a docker image of the HEA Registry Microservice.
        :param other_docker_images: an iterable of DockerContainerConfigs for docker images of other microservices.
        :param package_name: the service's distribution package name. If None, service metadata will not be registered
        with the registry service, which may be what you want if no registry service is running.
        """
        super().__init__(methodName=methodName, port=port)
        if href is None:
            href_ = f'/{coll}/'  # FIXME: href default needs to include root (http://localhost:...)
        else:
            href_ = str(href)
            if not href_.endswith('/'):
                href_ += '/'
        self._href = URL(href_)
        self._coll = str(coll)
        self._body_post = deepcopy(body_post)
        self._body_put = deepcopy(body_put)
        self._content_id = content_id
        self._expected_all = deepcopy(expected_all)
        self._expected_one = deepcopy(expected_one or expected_all)
        self._expected_one_wstl = deepcopy(expected_one_wstl or expected_all_wstl)
        self._expected_all_wstl = deepcopy(expected_all_wstl)
        self._expected_one_duplicate_form = deepcopy(expected_one_duplicate_form)
        self._expected_opener = expected_opener
        self._expected_opener_body = deepcopy(expected_opener_body)
        self._wstl_package = wstl_package
        validate_collection_keys(desktop_objects)
        self.__desktop_objects = deepcopy(convert_to_collection_keys(desktop_objects))
        if content is not None:
            validate_collection_keys(content)
        self._content = deepcopy(convert_to_collection_keys(content) if content else None)
        self._content_type = content_type
        self._put_content_status = put_content_status
        self._headers = {SUB: sub if sub is not None else NONE_USER, hdrs.X_FORWARDED_HOST: 'localhost:8080'}
        self.__registry_docker_image = registry_docker_image
        self.__other_docker_images = deepcopy(other_docker_images)
        self.__db_manager_cls = MockMongoManager if db_manager_cls is None else db_manager_cls
        self.__package_name = str(package_name) if package_name is not None else None
        self.maxDiff = None

    def run(self, result=None):
        """
        Runs a test using a freshly created MongoDB Docker container. The container is destroyed upon concluding
        the test.

        :param result: a TestResult object into which the test's result is collected.
        :return: the TestResult object.
        """
        _logger = logging.getLogger(__name__)
        with self._caplog.at_level(logging.DEBUG):
            with maybe_docker_is_not_running():
                with app_context(package_name=self.__package_name,
                                db_manager_cls=self.__db_manager_cls,
                                desktop_objects=self.__desktop_objects,
                                other_docker_images=list(
                                    self.__other_docker_images) if self.__other_docker_images is not None else None,
                                registry_docker_image=self.__registry_docker_image,
                                content=self._content,
                                wstl_builder_factory=wstl.builder_factory(self._wstl_package,
                                                                        href=self._href)) as (
                app, docker_container_ports):
                    self._docker_container_ports = docker_container_ports
                    self.__app = app
                    return super().run(result)


    async def get_application(self) -> Application:
        return self.__app

    def expected_one_id(self) -> str | None:
        """
        Gets the ID of the expected GET data.

        :return: an id string or None if not found.
        :raises TypeError: if _expected_one is None.
        """
        first_expected_one_data = self._expected_one[0]['collection']['items'][0]['data']
        return next((e.get('value') for e in first_expected_one_data if e['name'] == 'id'), None)

    def body_put_id(self) -> str | None:
        """
        Gets the ID of the target of a PUT request, determined by the value of the template ``_body_put``.

        :return: an id string or None if not found.
        :raises TypeError: if _body_put is None.
        """
        logging.getLogger(__name__).debug('Template is %s', self._body_put)
        return next((e.get('value') for e in self._body_put['template']['data'] if e['name'] == 'id'), None)


def get_test_case_cls_default(coll: str,
                              fixtures: Mapping[str, list[DesktopObjectDict]] | Mapping[
                                  CollectionKey, list[DesktopObjectDict]],
                              duplicate_action_name: str | None = None,
                              db_manager_cls: Type[MicroserviceDatabaseManager] = MockMongoManager,
                              wstl_package: Optional[str] = None,
                              content: Mapping[str, dict[str, bytes]] | Mapping[
                                  CollectionKey, dict[str, bytes]] | None = None,
                              content_type: Optional[str] = None,
                              put_content_status: Optional[int] = None,
                              href: Optional[str | URL] = None,
                              get_actions: Optional[list[Action]] = None,
                              get_all_actions: Optional[list[Action]] = None,
                              expected_opener: Optional[Link] = None,
                              registry_docker_image: Optional[RegistryContainerConfig] = None,
                              other_docker_images: Iterable[DockerContainerConfig] | None = None,
                              port: Optional[int] = None,
                              sub = NONE_USER,
                              exclude: Optional[list[str]] = None,
                              duplicate_action_actions: list[Action] | None = None,
                              package_name: str | None = None) -> Type[MicroserviceTestCase]:
    """
    Create a test case class for testing a specific HEA microservice.

    :param coll: the name of the collection that the microservice uses (required).
    :param fixtures: HEA desktop objects to load into the database, as a map of collection keys ->
    list of desktop object dicts (required).
    :param duplicate_action_name: the name of the duplication action for this microservice (optional).
    :param db_manager_cls: the type of the database manager that will be used as the default for collection keys that
    are strings. Under typical circumstances, this should be used by the collection specified in coll. For Mongo unit
    tests, use MockMongoManager. For AWS unit tests, use MockS3ManagerWithMockMongo. For Mongo integration tests, use
    DockerMongoManager. For AWS integration tests, use MockS3Manager. Defaults to MockMongoManager.
    :param wstl_package: the name of the package containing the WeSTL data package, in standard module format
    (foo.bar).
    :param content: a mapping of collection keys -> HEA object IDs -> content that will be included with the
    objects inserted into the database. If None, objects will not have content.
    :param content_type: the MIME type of the content included with the objects (e.g., text/plain; charset=utf-8).
    Must be specified if content_type is specified.
    :param put_content_status: the expected HTTP response code for updating the content of the HEA object. Normally
    should be 204 if the content is updatable and 405 if not. Default is None, which will cause associated tests to be
    skipped.
    :param href: the resource being tested. If None, uses http://localhost:{port}/{coll}/, where {coll} is the
    collection name, and {port} is the port ultimately used by this test case.
    :param get_actions: the list of actions associated with the GET one route of the microservice.
    :param get_all_actions: the list of actions associated with the GET all route of the microservice.
    :param expected_opener: The expected URL of the resource that does the opening, including a list of HTML rel
    values.
    :param registry_docker_image: a RegistryContainerConfig for a Docker image of the HEA Registry Microservice. It
    will be launched once the MongoDB container is live. If the microservice depends on there being a registry, you
    should use testenv.MockRegistryContainerConfig for unit tests and dockermongo.RealRegistryContainerConfig for
    integration tests. Must be specified if other_docker_images is specified.
    :param other_docker_images: an iterable of DockerContainerConfigs for Docker images of other microservices. They
    will be launched once the MongoDB container is live.
    :param port: the port number to run aiohttp. If None, a random available port will be chosen.
    :param sub: the user from which the requests will be sent. It should be set to heaobject.user.TEST_USER for
    permissions testing. Defaults to heaobject.user.NONE_USER.
    :param exclude: a list of expected value names to exclude from test case class creation. Tests that need any
    expected value you specify here will be skipped.
    :param package_name: the service's distribution package name, used to register metadata about the running service
    with the registry service.
    :return: A test case class for testing the microservice.
    """
    exclude_ = exclude if exclude is not None else []
    coll_ = coll if isinstance(coll, CollectionKey) else CollectionKey(name=coll, db_manager_cls=db_manager_cls)
    expected_values_ = {k: v for k, v in expected_values(fixtures, coll_, wstl.builder(package=wstl_package),
                                                         duplicate_action_name, href,
                                                         get_actions=get_actions,
                                                         get_all_actions=get_all_actions,
                                                         opener_link=expected_opener,
                                                         duplicate_action_actions=duplicate_action_actions,
                                                         exclude=exclude_,
                                                         sub=sub).items() if v is not None}

    class ExpectedValuesMicroserviceTestCase(MicroserviceTestCase):
        def __init__(self, methodName: str = 'runTest') -> None:
            super().__init__(package_name=package_name, coll=coll, desktop_objects=fixtures,
                             db_manager_cls=db_manager_cls,
                             wstl_package=wstl_package, href=href, content=content, content_type=content_type,
                             put_content_status=put_content_status, registry_docker_image=registry_docker_image,
                             other_docker_images=other_docker_images, port=port, sub=sub, methodName=methodName,
                             **expected_values_)

    return ExpectedValuesMicroserviceTestCase



"""
Constants for microservice defaults.
"""
from yarl import URL
import logging

DEFAULT_LOG_LEVEL = logging.DEBUG
DEFAULT_PORT = 8080
DEFAULT_BASE_URL = URL(f'http://localhost:{DEFAULT_PORT}')
DEFAULT_REGISTRY_URL = 'http://localhost:8080'



import unittest
import sys
from heaserver.service.testcase import testenv, mockaws, mockmongo

class MyTestCase(unittest.TestCase):
    def test_docker_imported(self):
        self.assertFalse('docker' in sys.modules)

    def test_testcontainers_imported(self):
        self.assertFalse('testcontainers' in sys.modules)
